unit PortFont;
(******************************************************************************
unit PortFont -- TPortableFont & TPortableFontComp

PURPOSE:

    TPortableFont lets you specify a TrueType font name. The font itself will
    be stored inside the EXE. If you run your EXE on a computer that does not
    have this font installed, TPortableFont can install it for you. You can
    specify whether the font should be installed temporarily or permanantly.

    TPortableFont is not a component -- it's a "persistent" class. Think of it
    like a TFont. It is usually a property in a component, not the component
    itself. You would use this when building other components.

    TPortableFontComp is a wrapper component whose only property is the
    TPortableFont. Use this if all you need is the TPortableFont on your form.

AUTHOR:

    Dan Thomas
    Apollo Enterprises
    dthomas@apollorx.com

DISCLAIMER:

    Use this at your own risk.

RIGHTS:

    Anyone can use this for anything they want. You don't have to mention me.

ENVIRONMENT:

    This has been tested under Delphi 2.01.

HISTORY:

    First release: April 22, 1998

    April 23, 1998
        Added the ability to create a Resource Script file (RC) from a font. See
        WriteToRcFile and then CreateFromRcData.

        Added AutoInstall property

        Fixed bug in how RegisterPropertyEditor was called that caused the
        property editor to be invoked on strings other than the FontName string.

MODIFICATIONS:

    If you keep me posted on any fixes, I'll try to maintain them.
*******************************************************************************)

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  DsgnIntf;

type

  TPortableFontName=string;

  TPortableFont=class(TPersistent)
    protected
      FFontName                : TPortableFontName;
      FOwner                   : TComponent;
      FFontFileID,
      FInstalledFontFileID,
      FInstalledFontResFileID  : string;
      FFontLoaded              : boolean;
      FData                    : pointer;
      FSize                    : longint;

      procedure SetFontName(Value: TPortableFontName);
      procedure ReleaseMemory;
      procedure DefineProperties(Filer: TFiler); override;
      procedure ReadBinaryData(Stream: TStream);
      procedure WriteBinaryData(Stream: TStream);
      procedure LoadFromFile(const sFileID: string);
      procedure WriteToFile(const sFileID: string);

    public
      constructor Create(AOwner: TComponent);
      constructor CreateFromRCData(AOwner: TComponent; RCDataID,AFontName,AFontFileID: string);
      destructor Destroy; override;

      procedure Assign(Source: TPersistent); override;

      procedure InstallFont;
      procedure RemoveFont;


      {This will create a resource compiler script (.rc) with one entry -- an
       RCDATA entry for this font, which you can then use BRCC32 to compile into
       a .res file.

       One use for this would be to use the $R compiler directive to include the
       .res file in your program then use CreateFromRCData to construct the
       object at run-time}
      procedure WriteToRcFile(sFileID,sIdentifier: string);

      property FontLoaded: boolean read FFontLoaded;

    published
      property FontName: TPortableFontName read FFontName write SetFontName;
  end; {TPortableFont}


  TPortableFontComp=class(TComponent)
    protected
      FPortableFont : TPortableFont;
      FAutoInstall  : boolean;

      procedure Loaded; override;
      procedure SetPortableFont(Value: TPortableFont);

    public
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;

    published
      property AutoInstall: boolean read FAutoInstall write FAutoInstall;
      property PortableFont: TPortableFont read FPortableFont write SetPortableFont;
  end; {TPortableFontComp}


  { This is a Property class. It allows us to provide a drop-down list that the
    user can use to choose the font name. }
  TPortableFontNameProperty=class(TStringProperty)
    public
      function GetAttributes: TPropertyAttributes; override;
      procedure GetValues(Proc : TGetStrProc); override;
  end; {TPortableFontNameProperty}




procedure Register;

implementation

uses Registry;

var
  {List of names of the TrueType fonts in the system}
  TrueTypeFontNames : TStringList;

{$R *.res}






{******************************************************************************
Routines for the TrueTypeFontNames list
*******************************************************************************}

function EnumTrueTypeFontsProc(var LogFont    : TLogFont;
                               var TextMetric : TTextMetric;
                               FontType       : Integer;
                               Data           : Pointer): Integer; stdcall;

begin
  {only TrueType fonts}
  if TextMetric.tmPitchAndFamily and TMPF_TRUETYPE = TMPF_TRUETYPE then
    TStringList(Data).Add(StrPas(LogFont.lfFaceName));
  Result := 1;
end; {EnumTrueTypeFontsProc}

procedure RefreshTrueTypeFonts;
var
  DC : HDC;
begin
  TrueTypeFontNames.Clear;
  DC := GetDC(0);
  try
    EnumFontFamilies(DC,nil,@EnumTrueTypeFontsProc,integer(TrueTypeFontNames));
  finally
    ReleaseDC(0, DC);
  end;
end; {RefreshTrueTypeFonts}

procedure InitTrueTypeFonts;
begin
  TrueTypeFontNames := TStringList.Create;
  RefreshTrueTypeFonts;
end; {InitTrueTypeFonts}








{******************************************************************************
TPortableFont
*******************************************************************************}

constructor TPortableFont.Create(AOwner: TComponent);
begin
  inherited Create;

  {we need an Owner so we can check the ComponentState, since TPersistent does
   not get a ComponentState property}
  FOwner := AOwner;
end; {TPortableFont.Create}


constructor TPortableFont.CreateFromRCData(AOwner: TComponent; RCDataID,AFontName,AFontFileID: string);
var
  HResInfo : HRSRC;
  HGlobal  : THandle;
  p        : pointer;
  n        : longint;
begin
  {this loads the font from an RCDATA resource included in the exe -- see the
   comments at the start of this unit regarding WriteToRcFile}

  inherited Create;
  FOwner := AOwner;

  HResInfo := FindResource(HInstance,PChar(RCDataID),RT_RCDATA);
  if HResInfo = 0 then
    raise Exception.Create('Could not find RCData for '+RCDataID);

  try
    HGlobal := LoadResource(HInstance,HResInfo);
    if HGlobal = 0 then
      raise Exception.Create('Could not load resource '+RCDataID);

    try
      p := LockResource(HGlobal);
      n := SizeOfResource(HInstance,HResInfo);
      GetMem(FData,n);
      if FData = nil then
        raise Exception.Create('Out of memory');
      FSize := n;
      move(p^,FData^,FSize);
    finally
      UnlockResource(HGlobal);
    end; {try..finally}

  finally
    FreeResource(HResInfo);
  end; {try..finally}

  FFontName :=AFontName;
  FFontFileID := AFontFileID;
end; {TPortableFont.CreateFromRCData}


destructor TPortableFont.Destroy;
begin
  {clean up any installed fonts and any memory still allocated}
  RemoveFont;
  ReleaseMemory;

  inherited Destroy;
end; {TPortableFont.Destroy}


procedure TPortableFont.DefineProperties(Filer: TFiler);
begin
  {this procedure allows us to add our "ReadBinaryData" and "WriteBinaryData"
   procedures to the streaming process}

  inherited DefineProperties(Filer);
  Filer.DefineBinaryProperty('PortableFont',ReadBinaryData,WriteBinaryData,true);
end; {TPortableFont.DefineProperties}


procedure TPortableFont.ReleaseMemory;
begin
  if FData <> nil then
  begin
    FreeMem(FData);
    FData := nil;
    FSize := 0;
  end;
end; {TPortableFont.ReleaseMemory}


procedure TPortableFont.ReadBinaryData(Stream: TStream);
var
  n : longint;

  function ReadString: shortstring;
  begin
    Stream.ReadBuffer(Result[0],1);
    if Result <> '' then
      Stream.ReadBuffer(Result[1],ord(Result[0]));
  end; {ReadString}

begin {TPortableFont.ReadBinaryData}
  {this gets called when we're being read from a stream (i.e. the .DFM file at
   design time, or the compiled EXE at run time}

  {this is the first thing we stored on the stream}
  FFontFileID := ReadString;

  {make sure we clean up if needed (probably not needed, but just in case)}
  RemoveFont;
  ReleaseMemory;

  {read the size of the data}
  Stream.ReadBuffer(n,SizeOf(n));

  {if there's any data to read, allocate the memory and read it}
  if n > 0 then
  begin
    GetMem(FData,n);
    if FData = nil then
      raise Exception.Create('Out of memory');
    FSize := n;
    Stream.ReadBuffer(FData^,n);
  end;
end; {TPortableFont.ReadBinaryData}


procedure TPortableFont.WriteBinaryData(Stream: TStream);

  procedure WriteString(s: shortstring);
  begin
    Stream.WriteBuffer(s[0],1);
    if s <> '' then
      Stream.WriteBuffer(s[1],ord(s[0]));
  end; {WriteString}

begin {TPortableFont.WriteBinaryData}
  {this gets called when we're being written to the DFM file}

  {save this first}
  WriteString(FFontFileID);

  {write the length of the data}
  Stream.WriteBuffer(FSize,sizeof(FSize));

  {if there's any data, write it also}
  if FSize > 0 then
    Stream.WriteBuffer(FData^,FSize);
end; {TPortableFont.WriteBinaryData}


procedure TPortableFont.Assign(Source: TPersistent);
begin
  if Source is TPortableFont then
  begin
    FontName := TPortableFont(Source).FontName;
    Exit;
  end;

  {calling the ancestor will raise an exception saying that "Source" is not
   a valid object type for this Assign}
  inherited Assign(Source);
end; {TPortableFont.Assign}


procedure TPortableFont.LoadFromFile(const sFileID: string);
var
  hFile     : integer;
  nFileSize : longint;
begin
  {this procedure loads the data from a TTF file into memory -- it should only
   happen at design time, but I don't check for that}

  {clean up any existing font}
  RemoveFont;
  ReleaseMemory;

  {try to open the specified file}
  hFile := FileOpen(sFileID,fmShareDenyWrite);
  if hFile < 0 then
    raise Exception.Create('Could not open file '+sFileID);

  try

    {this gets us the size of the file}
    nFileSize := FileSeek(hFile,0,2);

    {allocate the memory}
    GetMem(FData,nFileSize);
    if FData = nil then
      raise Exception.Create('File '+sFileID+' is too large, or out of memory');

    try
      {store the size of the data}
      FSize := nFileSize;

      {go back to the start of the file}
      FileSeek(hFile,0,0);

      {read the data into memory}
      if FileRead(hFile,FData^,FSize) <> FSize then
        raise Exception.Create('Error reading file '+sFileID);

    except
      on exception do
      begin
        ReleaseMemory;
        raise;
      end;
    end; {try..except}

  finally
    FileClose(hFile);
  end; {try..finally}
end; {TPortableFont.LoadFromFile}


procedure TPortableFont.WriteToFile(const sFileID: string);
var
  hFile : integer;
begin
  {this gets called from InstallFont}

  if FData = nil then
    raise Exception.Create('No data to write');

  {try to create the file}
  hFile := FileCreate(sFileID);
  if hFile < 0 then
    raise Exception.Create('Could not create file '+sFileID);

  try
    {try to write the data}
    if FileWrite(hFile,FData^,FSize) <> FSize then
      raise Exception.Create('Error writing to file '+sFileID);

  finally
    FileClose(hFile);
  end; {try..finally}
end; {TPortableFont.WriteToFile}


procedure TPortableFont.WriteToRcFile(sFileID,sIdentifier: string);
type
  PArray=^TArray;
  TArray=array[0..maxint-1] of byte;
var
  fOut         : TextFile;
  nBufferIndex,
  nByteCount,
  nSize        : integer;
  s            : shortstring;

  procedure FinishLine;
  begin
    if nByteCount > 0 then
    begin
      writeln(fOut,s+'''');
      s := '';
      nByteCount := 0;
    end;
  end; {FinishLine}

begin {TPortableFont.WriteToRcFile}
  {writes the font data to an RCDATA resource file -- see comments at the top of
   this unit for more information}
   
  if FData = nil then
    raise Exception.Create('No data to write to RC file');

  assignfile(fOut,sFileID);
  rewrite(fOut);
  try
    writeln(fOut,sIdentifier+' RCDATA');
    writeln(fOut,'BEGIN');

    s := '';
    nByteCount := 0;
    for nBufferIndex := 0 to FSize do
    begin
      if s = '' then
        s := ''''+IntToHex(PArray(FData)^[nBufferIndex],2)
      else
        s := s+' '+IntToHex(PArray(FData)^[nBufferIndex],2);

      inc(nByteCount);
      if nByteCount >= 16 then
        FinishLine;
    end; {of while}

    FinishLine;

    writeln(fOut,'END');

  finally
    closefile(fOut);
  end; {try..finally}
end; {TPortableFont.WriteToRcFile}


procedure TPortableFont.InstallFont;
var
  sTempDir  : shortstring;
  sRootName : string;
begin
  {install the font, if any font is specified}

  {if there's not font info to install, just exit}
  if (FFontLoaded) or (FFontName = '') or (FData = nil) then
    exit;

  {see if the font is already installed}
  RefreshTrueTypeFonts;
  if TrueTypeFontNames.IndexOf(FFontName) >= 0 then
  begin
    {it is, so realse our memory and exit}
    if not (csDesigning in FOwner.ComponentState) then
      ReleaseMemory;
    exit;
  end;

  {get the temp dir -- this is kind of kludgey because I don't know if
   GetTempPath can ever fail -- I'm assuming it can't}
  sTempDir[0] := char(GetTempPath(sizeof(sTempDir)-2,@sTempDir[1]));
  if sTempDir = '' then
    exit;
  if sTempDir[length(sTempDir)] <> '\' then
    sTempDir := sTempDir+'\';

  {the name we'll use is based on the original name of the font file}
  sRootName := ExtractFileName(FFontFileID);

  {we need both a resource (.fot) file and a ttf file}
  FInstalledFontResFileID := ChangeFileExt(sTempDir+sRootName,'.fot');
  FInstalledFontFileID := ChangeFileExt(sTempDir+sRootName,'.ttf');

  {save our data to disk}
  WriteToFile(FInstalledFontFileID);

  {we don't need the data in memory any more}
  if not (csDesigning in FOwner.ComponentState) then
    ReleaseMemory;

  {try to create the resource file -- I tried using a parameter of "1" here for
   a temporary font, but it didn't work -- anyone know why?}
  CreateScalableFontResource(0,PChar(FInstalledFontResFileID),PChar(FInstalledFontFileID),nil);

  {did we succeed?}
  if FileExists(FInstalledFontResFileID) then
  begin

    {try to add the font to the system}
    if AddFontResource(PChar(FInstalledFontResFileID)) <> 0 then
    begin
      {we added the font}
      FFontLoaded := true;

      {this is what the docs say we need to do}
      SendMessage(HWND_BROADCAST,WM_FONTCHANGE,0,0);
    end;
  end;

  {note that no provision has been made for being unsuccessful -- if you need
   something, add it}
end; {TPortableFont.InstallFont}


procedure TPortableFont.RemoveFont;
begin
  {remove the font if needed}
  if FFontLoaded then
  begin
    {remove it from the system}
    RemoveFontResource(PChar(FInstalledFontResFileID));
    FFontLoaded := false;

    {the docs tell us to do this}
    SendMessage(HWND_BROADCAST,WM_FONTCHANGE,0,0);

    {delete the files}
    DeleteFile(FInstalledFontResFileID);
    DeleteFile(FInstalledFontFileID);
  end;
end; {TPortableFont.RemoveFont}


procedure TPortableFont.SetFontName(Value: TPortableFontName);
var
  reg     : TRegIniFile;
  sDir,
  sFileID : shortstring;
begin
  {this gets called when the FontName property changes}

  {if we're loading, then just save the value of the name and exit}
  if csLoading in FOwner.ComponentState then
  begin
    FFontName := Value;
    exit;
  end;

  {we should never get here at run-time}
  if not (csDesigning in FOwner.ComponentState) then
    raise Exception.Create('FontName changed at run-time');

  {clean up stuff}
  RemoveFont;
  ReleaseMemory;

  {if we abort out of this procedure, these are the defaults}
  FFontName := '';
  FFontFileID := '';

  {if the new value is an empty string, then we're out of here}
  if Value = '' then
    exit;


  {if we get here, we want to try to read the font file into memory --
   design-time only}


  reg := TRegIniFile.Create('');
  try
    {get the file ID}
    reg.RootKey := HKEY_LOCAL_MACHINE;
    sFileID := reg.ReadString('\Software\Microsoft\Windows\CurrentVersion\Fonts',Value+' (TrueType)','');
    if sFileID = '' then
      raise Exception.Create('Could not find registry entry for FontName');

    {get the Fonts dir}
    reg.RootKey := HKEY_USERS;
    sDir := reg.ReadString('\.Default\Software\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders','Fonts','');
    if sDir = '' then
    begin
      sDir[0] := char(GetWindowsDirectory(@sDir[1],sizeof(sDir)-2));
      if (sDir = '') or (sDir[length(sDir)] <> '\') then
        sDir := sDir+'\';
      sDir := sDir+'Fonts\';
    end
    else
    if sDir[length(sDir)] <> '\' then
      sDir := sDir+'\';
  finally
    reg.Free;
  end; {try..finally}

  {try to load the font file into memory}
  LoadFromFile(sDir+sFileID);

  {we've apparently succeeded, so store the names}
  FFontName := Value;
  FFontFileID := sDir+sFileID;
end; {TPortableFont.SetFontName}






{******************************************************************************
TPortableFontComp
*******************************************************************************}

constructor TPortableFontComp.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FPortableFont := TPortableFont.Create(self);
  FAutoInstall := true;
end; {TPortableFontComp.Create}

destructor TPortableFontComp.Destroy;
begin
  FPortableFont.Free;
  inherited Destroy;
end; {TPortableFontComp.Destroy}

procedure TPortableFontComp.Loaded;
begin
  inherited Loaded;

  {note: we call this regardless of whether this is design time or tum-time, so
   if you're loading a .DFM in Delphi that has one of these with a font you
   don't have on your system, you'll get it}
  if FAutoInstall then
    FPortableFont.InstallFont;
end; {TPortableFontComp.Loaded}

procedure TPortableFontComp.SetPortableFont(Value: TPortableFont);
begin
  FPortableFont.Assign(Value);
end; {TPortableFontComp.SetPortableFont}





{******************************************************************************
TPortableFontNameProperty
*******************************************************************************}

function TPortableFontNameProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paValueList, paSortList, paReadOnly];
end; {TPortableFontNameProperty.GetAttributes}

procedure TPortableFontNameProperty.GetValues(Proc : TGetStrProc);
var
  x : integer;
begin
  for x := 0 to TrueTypeFontNames.Count-1 do
    Proc(TrueTypeFontNames[x]);
end; {TPortableFontNameProperty.GetValues}






procedure Register;
begin
  RegisterComponents('Apollo', [TPortableFontComp]);
  RegisterPropertyEditor(TypeInfo(TPortableFontName),TPortableFont,'FontName',TPortableFontNameProperty);
end;








initialization
  InitTrueTypeFonts;
finalization
  TrueTypeFontNames.Free;
end.
